from abc import ABC
from typing import Iterable


class BaseField(ABC):

    def __init__(self, value):
        self.value = value


class UniqueField(BaseField):

    def __init__(self, value, autoincrement=True):
        super().__init__(value)
        self.autoincrement = autoincrement


class IterableField(BaseField):

    def __init__(self, value: Iterable, iterative_mode="ORT"):
        """
        iterative_mode: `ORT` - Orthogonal
                        `EXH` - Exhaustive
        """
        super().__init__(value)
        self.iterative_mode = iterative_mode


class FixedField(BaseField):
    pass



# -*- coding: utf-8 -*-
"""
# @Time    : 2022/3/5 14:25
# @Author  : bruce
# @desc    :
"""
import sys

from api_test_ez.ez import get_config, Log
from api_test_ez.ez.decorator import singleton
from api_test_ez.project.configs import Configs
from api_test_ez.project.settings import Settings
import os

ENV_EZ_PROJECT_DIR = 'EZ_PROJECT_DIR'
ENV_EZ_SETTINGS_MODULE = 'EZ_SETTINGS_MODULE'


def closest_file(file_name, path='.', prev_path=None):
    """Return the path to the closest project.cfg file by traversing the current
    directory and its parents
    """
    if path == prev_path:
        return ''
    path = os.path.abspath(path)
    cfg_file = os.path.join(path, file_name)
    if os.path.exists(cfg_file):
        return cfg_file
    return closest_file(file_name, os.path.dirname(path), path)


def search_file(file_name: str, path: str = '.', prev_path: str = None, search_result=None):
    """Return the path list to the case.cfg file by traversing the current
    directory and its parents
    """
    if search_result is None:
        search_result = []
    if path == prev_path:
        return search_result
    path = os.path.abspath(path)
    cfg_file = os.path.join(path, file_name)
    if os.path.exists(cfg_file):
        search_result.append(cfg_file)
    return search_file(file_name, path=os.path.dirname(path), prev_path=path, search_result=search_result)


def init_project(project='default'):
    # project config
    ez_cfg_name = 'project.cfg'
    project_cfg_path = closest_file(ez_cfg_name)
    cfg = get_config(project_cfg_path)
    if cfg.has_option('settings', project):
        project_settings_path = cfg.get('settings', project)
        os.environ[ENV_EZ_SETTINGS_MODULE] = project_settings_path
        os.environ[ENV_EZ_PROJECT_DIR] = os.path.abspath(os.path.dirname(project_cfg_path))
        sys.path.append(os.environ[ENV_EZ_PROJECT_DIR])


def get_ez_settings():
    if ENV_EZ_SETTINGS_MODULE not in os.environ:
        init_project()

    _settings = Settings()
    settings_module_path = os.environ.get(ENV_EZ_SETTINGS_MODULE)
    if settings_module_path:
        _settings.set_module(settings_module_path)

    return _settings


def get_ez_config(ez_file_path):
    # ez case config
    ez_cfg_name = 'ez.cfg'
    ez_cfg_filelist = search_file(ez_cfg_name, path=ez_file_path)

    _configs = Configs()
    support_file_config_priority = ['package', 'module', 'project']
    for ez_cfg_file in ez_cfg_filelist:
        if len(support_file_config_priority) <= 0:
            break
        _configs.set_config(ez_cfg_file, support_file_config_priority.pop(0))

    return _configs


def get_ez_logger(project_settings: Settings, logger_name=None):
    """
    :param project_settings:
    :param logger_name:
    :return:
    """
    console_format = project_settings.get('CONSOLE_LOG_FORMAT')
    console_log_level = project_settings.get('CONSOLE_LOG_LEVEL')

    file_format = project_settings.get('FILE_LOG_FORMAT')
    file_log_level = project_settings.get('FILE_LOG_LEVEL')
    file_log_path = project_settings.get('FILE_LOG_PATH')

    log = Log(logger_name=logger_name)

    if console_format:
        log.console_format = console_format

    if console_log_level:
        log.console_log_level = console_log_level

    if file_format:
        log.file_format = file_format

    if file_log_level:
        log.file_log_level = file_log_level

    logger = log.init_logger(file_log_path)

    return logger


class ReportConfig:

    def __init__(self, project_settings: Settings):
        self._report_dir = project_settings.get("REPORT_DIR")
        self._report_file_name = project_settings.get("REPORT_FILE_NAME")

        self._report_file_path = os.path.join(self._report_dir, self._report_file_name) if self._report_dir else None
        self._report_title = project_settings.get("REPORT_TITLE")
        self._report_desc = project_settings.get("REPORT_DESC")
        self._report_theme = project_settings.get("BR_REPORT_THEME")

    @property
    def report_file_path(self):
        return self._report_file_path

    @property
    def report_title(self):
        return self._report_title

    @property
    def report_desc(self):
        return self._report_desc

    @property
    def report_dir(self):
        return self._report_dir

    @property
    def report_file_name(self):
        return self._report_file_name

    @property
    def theme(self):
        return self._report_theme


@singleton
class Project:
    """Ez initialize"""
    def __init__(self, ez_file_path, env_name=None):
        self.env_name = env_name if env_name else os.path.dirname(ez_file_path)
        self._settings = get_ez_settings()
        self._configs = get_ez_config(ez_file_path)
        self._logger = get_ez_logger(self._settings, self.env_name)
        # self._report = ReportConfig(self._settings)

    @property
    def settings(self):
        return self._settings

    @property
    def configs(self):
        return self._configs

    @property
    def logger(self):
        return self._logger

    @property
    def report(self):
        # We should always get the newest report config.
        return ReportConfig(self._settings)



# -*- coding: utf-8 -*-
"""
# @Time    : 2022/3/5 13:59
# @Author  : bruce
# @desc    :
"""
import argparse
import os
import pathlib
import sys
from unittest import TestProgram

from api_test_ez.core.report import BRReporter, HtmlReporter, DryRun
from api_test_ez.project import Project

EZ_SUPPORT_ACTION = [
    'run',
    'dry-run'
]


class EzCommand:

    def __init__(self, argv=None):
        if argv is None:
            argv = sys.argv[1:]
        self.project = None
        self.process_args(argv)

    @staticmethod
    def parse_ez_parent():
        parser = argparse.ArgumentParser(add_help=False, exit_on_error=False)
        # action
        parser.add_argument('action',
                            choices=['run'],
                            help='Command action.'
                                 'For example: `ez run` <test_cases_path>')
        # run cases
        parser.add_argument('cases_path',
                            help='Run the next parameters as the test case.'
                                 'For example: ez run <test_cases_path>')
        return parser

    @staticmethod
    def parse_ez_args():
        # parser = argparse.ArgumentParser(add_help=False, parents=[parent])
        parser = argparse.ArgumentParser(add_help=True, exit_on_error=False)
        # action
        parser.add_argument('action',
                            choices=['run', 'dry-run'],
                            help='Command action.'
                                 'For example: `ez run`')
        # run cases
        parser.add_argument('cases_path',
                            help='Run the next parameters as the test case.'
                                 'For example: ez run <test_cases_path>')
        # version
        with open(os.path.join(os.path.dirname(__file__), 'VERSION'), 'rb') as f:
            version = f.read().decode('ascii').strip()
        parser.add_argument('-version', '--version',
                            action='version',
                            version=f'ApiTestEz version {version}',
                            help='Current EZ version.')
        # framework
        parser.add_argument('-fk', '--framework', dest='framework',
                            default='unittest',
                            choices=['unittest', 'pytest'],
                            help='`unittest` or `pytest`, how to EZ run cases, `unittest` as default.')
        # config
        parser.add_argument('-cfg', '--config', dest='config',
                            action="extend", nargs="+",
                            help='Set EZ config, priority `command`. '
                                 'For example: `-cfg host=127.0.0.1`, '
                                 'details: https://github.com/bruce4520196/ApiTestEz.')
        # config file
        parser.add_argument('-cfgf', '--config-file', dest='config_file',
                            help='Set EZ <config_file_path>, priority `command`. '
                                 'details: https://github.com/bruce4520196/ApiTestEz.')
        # report style
        parser.add_argument('-rs', '--report-style', dest='report_style',
                            default='br',
                            choices=['html', 'br'],
                            help='Report style. default `html`. '
                                 'support: `html` (ie: HtmlReporter), `br` (ie: BRReporter)')

        parser.add_argument('-t', '--tester', dest='tester',
                            default='Jenkins',
                            help='Tester name.')

        # beautiful report theme
        parser.add_argument('-rt', '--report-theme', dest='report_theme',
                            default='theme_default',
                            choices=['theme_default', 'theme_default', 'theme_cyan', 'theme_candy', 'theme_memories'],
                            help='Beautiful report theme. default `theme_default`. '
                                 'support: `theme_default`, `theme_default`,`theme_default`, '
                                 '`theme_cyan`, `theme_candy`, `theme_memories`')

        # report file
        parser.add_argument('-rf', '--report-file', dest='report_file',
                            help='Report file path.')
        return parser

    def process_args(self, argv):
        ez_parser = self.parse_ez_args()

        if len(argv) == 0:
            ez_parser.print_help()
            return

        if argv[0] == '-version' or argv[0] == '--version':
            ez_parser.parse_args()
            return

        if argv[0] not in EZ_SUPPORT_ACTION:
            print(
                f'EZ COMMAND ERROR: unknown action-word `{argv[0]}`. Support word must in {EZ_SUPPORT_ACTION}. See: \n')
            ez_parser.print_help()
            return

        if len(argv) == 1:
            print(f'EZ COMMAND ERROR: expect a <case-file-path> after the action-word `{argv[0]}`. See: \n')
            ez_parser.print_help()
            return

        if argv[1].startswith('-'):
            print(f'EZ COMMAND ERROR: expect a <case-file-path> after the action-word `{argv[0]}`, '
                  f'but `{argv[1]}` found. See: \n')
            ez_parser.print_help()
            return
        args, unknown_args = ez_parser.parse_known_args(argv)

        cases_path = args.cases_path
        if not cases_path:
            print('EZ COMMAND ERROR: `cases_path` not found. See: \n')
            ez_parser.print_help()
            return
        if not os.path.isfile(cases_path) and not os.path.isdir(cases_path):
            print('EZ COMMAND ERROR: `cases_path` is not a file or dir. See: \n')
            ez_parser.print_help()
            return
        project = Project(ez_file_path=cases_path if os.path.isdir(cases_path) else os.path.dirname(cases_path))

        if args.report_style:
            project.settings.set('REPORT_STYLE', args.report_style)

        if args.report_theme:
            project.settings.set('BR_REPORT_THEME', args.report_theme)

        if args.report_file:
            report_dir = os.path.dirname(args.report_file) \
                if os.path.dirname(args.report_file) \
                else os.path.dirname(__file__)
            project.settings.set('REPORT_DIR', report_dir)
            project.settings.set('REPORT_FILE_NAME', os.path.basename(args.report_file))

        if args.config_file:
            if os.path.exists(args.config_file):
                print(f'EZ COMMAND ERROR: can not find ez-config-file `{args.config_file}`.')
                ez_parser.print_help()
            else:
                project.configs.set_config(args.config_file, priority='command')

        if args.config:
            for cfg in args.config:
                if '=' not in cfg:
                    print('EZ COMMAND ERROR: config format error, For example: `-cfg host=127.0.0.1`. See: \n')
                    ez_parser.print_help()
                    return
                project.configs.set(*cfg.split('=', 1), priority='command')

        if args.action == 'run':
            if project.report.report_dir:
                if args.report_style == 'br':
                    BRReporter(args.cases_path, tester=args.tester).run()
                    return
                elif args.report_style == 'html':
                    HtmlReporter(args.cases_path, tester=args.tester).run()
                    return
                else:
                    print(f'EZ COMMAND WARNING: `{args.report_style}` is not supported `report_style`, '
                          f'run tests as dry-run. See: \n')
                    ez_parser.print_help()
            else:
                print(f'EZ COMMAND WARNING: `report_dir` does not set, run tests as dry-run.\n')
                # ez_parser.print_help()
            DryRun(args.cases_path).run()
            return

        elif args.action == 'dry-run':
            DryRun(args.cases_path).run()
            return

        else:
            print(f'EZ COMMAND ERROR: {args.action} is not supported `action`. See: \n')
            ez_parser.print_help()
            return


def main():
    EzCommand()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
# @Time    : 2022/6/9 19:30
# @Author  : bruce
# @desc    :
"""


class HttpRequestException(Exception):
    def __init__(self, err):
        Exception.__init__(self, err)


class CaseFileNotFoundException(Exception):
    def __init__(self, err):
        Exception.__init__(self, err)



import os

import tablib

from api_test_ez.core.case import CaseFileNotFoundException
from api_test_ez.project import ENV_EZ_PROJECT_DIR


class CaseLoaderMiddleware(object):

    def __init__(self, configs):
        # ez configs
        self.configs = configs

    def load_test_data(self) -> list:
        raise NotImplementedError


class FileCaseLoaderMiddleware(CaseLoaderMiddleware):

    def load_test_data(self) -> list:
        data_filename = self.configs.get("case_filepath")
        if not data_filename:
            return []
        if isinstance(data_filename, str):
            # 增加绝对路径和相对路径兼容
            if not os.path.exists(data_filename):
                data_filename = os.path.join(os.environ[ENV_EZ_PROJECT_DIR], data_filename)
                if not os.path.exists(data_filename):
                    raise CaseFileNotFoundException(
                        err=f'Case file "{data_filename}" not found from both relative and absolute paths.')
            with open(data_filename, 'rb') as f:
                data_set = tablib.Dataset().load(f.read())
                return data_set.dict
        else:
            raise CaseFileNotFoundException(
                err=f'Case filename type err: expect "str" but "{type(data_filename)}" found.')



import itertools

from collections import namedtuple
from typing import Dict, List, Iterable, Callable

import tablib
from allpairspy import AllPairs

from api_test_ez.ez.case_builder.fields import UniqueField, IterableField, BaseField, FixedField


def to_pairs(name, field):     # In order to be consistent with the `allpairspy.Pairs` type
    return namedtuple("Pairs", name)(field)


class CaseBuilderSchemaMeta(type):

    def __new__(mcs, name, bases, attrs):
        klass = super().__new__(mcs, name, bases, attrs)

        # If a base class just call super new
        if name == "CaseBuilderSchema":
            return klass

        # filter fields
        klass._ort_iterable_fields = []
        klass._exh_iterable_fields = []
        klass._unique_field = []
        klass._fixed_field = []

        for attr_name, attr_value in attrs.items():
            if issubclass(attr_value.__class__, IterableField):
                if attr_value.iterative_mode == 'ORT':
                    klass._ort_iterable_fields.append(
                        mcs._convert_field_to_list(attr_name, attr_value.value)
                    )
                elif attr_value.iterative_mode == 'EXH':
                    klass._exh_iterable_fields.append(
                        mcs._convert_field_to_list(attr_name, attr_value.value)
                    )
            elif issubclass(attr_value.__class__, UniqueField):
                klass._unique_field.append(
                       {attr_name: attr_value.value}
                    )
            elif issubclass(attr_value.__class__, FixedField):
                klass._fixed_field.append(
                       {attr_name: attr_value.value}
                    )
        return klass

    @staticmethod
    def _convert_field_to_list(
            field_name,
            field_value: Iterable
    ) -> List[dict]:
        return [{field_name: value} for value in field_value]


class CaseBuilderSchema(metaclass=CaseBuilderSchemaMeta):

    def __init__(
            self,
            ort_filter_func: Callable = lambda x: True,
            exh_filter_func: Callable = lambda x: True
    ):
        self._ort_filter_func = ort_filter_func
        self._exh_filter_func = exh_filter_func

    def _fields_build(self):

        # self._exh_iterable_fields:
        # [
        #   [{"a": 1}, {"a": 2}],
        #   [{"b": 3}, {"b": 4}],
        # ]

        # self._ort_iterable_fields:
        # [
        #   [{"c": 5}, {"c": 6}],
        #   [{"d": 7}, {"d": 8}],
        # ]

        # 1. generate the "EXH" (Exhaustive) field.
        computed_fields = []
        if len(self._exh_iterable_fields) > 0:
            # computed_exh_fields:
            # [
            #   [{"a": 1}, {"b": 3}],
            #   [{"a": 1}, {"b": 4}],
            #   [{"a": 2}, {"b": 3}],
            #   [{"a": 2}, {"b": 4}],
            # ]
            computed_fields = self._iter_field_build(self._exh_iterable_fields, "EXH")

        # 2. generate the "ORT" (Orthogonal) field.
        if len(self._ort_iterable_fields) > 0:
            if len(computed_fields) > 0:
                # `self._ort_iterable_fields` append `computed_exh_fields`:
                # [
                #   # self._ort_iterable_fields
                #   [{"c": 5}, {"c": 6}],
                #   [{"d": 7}, {"d": 8}],

                #   # computed_exh_fields
                #   [
                #       [{"a": 1}, {"b": 3}],
                #       [{"a": 1}, {"b": 4}],
                #       [{"a": 2}, {"b": 3}],
                #       [{"a": 2}, {"b": 4}],
                #   ],
                # ]
                self._ort_iterable_fields.append(computed_fields)
            computed_fields = self._iter_field_build(self._ort_iterable_fields, "ORT")

        # computed_ort_fields:
        # [
        #     [{'c': 5}, {'d': 7}, [{'a': 1}, {'b': 3}]],
        #     [{'c': 6}, {'d': 8}, [{'a': 1}, {'b': 3}]],
        #     [{'c': 6}, {'d': 7}, [{'a': 1}, {'b': 4}]],
        #     [{'c': 5}, {'d': 8}, [{'a': 1}, {'b': 4}]],
        #     [{'c': 5}, {'d': 8}, [{'a': 2}, {'b': 3}]],
        #     [{'c': 6}, {'d': 7}, [{'a': 2}, {'b': 3}]],
        #     [{'c': 6}, {'d': 7}, [{'a': 2}, {'b': 4}]],
        #     [{'c': 5}, {'d': 8}, [{'a': 2}, {'b': 4}]],
        # ]

        # self._unique_field
        # [{"e": 9}]

        # self._fixed_field
        # [{"f": 10}, {"g": 11}]

        # 3. merge `computed_ort_fields` / `fixed_field` / `unique_field`.
        merged_rows = []
        if len(computed_fields) > 0:
            for i, row in enumerate(computed_fields):
                # merge `fixed_field`
                new_row = row + self._fixed_field

                # merge `unique_field`
                for unique_field in self._unique_field:
                    new_unique_field = {
                        name: f"{value}_{str(i).zfill(len(str(len(computed_fields))))}"
                        for name, value in unique_field.items()
                    }

                    new_row.append(new_unique_field)

                merged_rows.append(new_row)
        else:   # maby `computed_ort_fields` is []
            # merge `fixed_field`
            new_row = self._fixed_field

            # merge `unique_field`
            for unique_field in self._unique_field:
                new_unique_field = {
                    name: f"{value}_01"
                    for name, value in unique_field.items()
                }

                new_row.append(new_unique_field)

            merged_rows.append(new_row)

        return merged_rows

    def unpack_row(self):
        # rows:
        # [{'a': CaseBuilderSchema}, {'b': '2'}, [{'c': '6'}, {'d': '8'}], {'name': 'case_23'}],
        rows = self._fields_build()
        new_rows = []
        for row in rows:
            new_row = []
            inner_row_schema = None
            for cell in row:
                if isinstance(cell, list):
                    # [{'c': '4'}, {'d': '7'}]
                    new_row += cell
                elif isinstance(cell, dict):
                    for v in cell.values():
                        try:
                            if issubclass(v, CaseBuilderSchema):
                                inner_row_schema = v
                        except TypeError:
                            if isinstance(v, list):
                                new_row += v
                            else:
                                new_row.append(cell)

                else:
                    raise Exception(f'Unknown type in row: {cell}')

            if inner_row_schema:
                # Exhaustive `new_row` and `inner_row_list`
                inner_rows = inner_row_schema().unpack_row()

                # `inner_rows`: [[{'name': '1'}, {'age': 2}]]
                # `new_row`: [{'b': '2'}, {'c': '4'}, {'d': '7'}, {'fix': 5}, {'name': 'case_01'}]
                # merge `inner_rows` and `new_row`
                for inner_row in inner_rows:
                    new_rows.append(inner_row + new_row)
            else:
                new_rows.append(new_row)
        return new_rows

    def _iter_field_build(self, iterable_fields, rule):
        if len(iterable_fields) <= 1:
            print("Warning: Algorithm-Rule must provide more than one option.")
            return iterable_fields

        if rule == "EXH":
            return self._exh_algorithm(iterable_fields, filter_func=self._exh_filter_func)
        elif rule == "ORT":
            return self._ort_algorithm(iterable_fields, filter_func=self._ort_filter_func)
        else:
            raise Exception(f"Unknown Algorithm-Rule: {rule}")

    @staticmethod
    def _ort_algorithm(iterable_fields: List[List[dict]], filter_func: Callable = lambda x: True):
        return [pair for pair in AllPairs(iterable_fields, filter_func=filter_func)]

    @staticmethod
    def _exh_algorithm(iterable_fields: List[List[dict]], filter_func: Callable = lambda x: True) -> List[List[dict]]:
        _exh_pairs = []
        for pair in itertools.product(*iterable_fields):

            if filter_func(pair):
                _exh_pairs.append(list(pair))

        return _exh_pairs

    def build(self):
        """
        :rtype: [ [dict] ]
        """
        return self.unpack_row()

    def save(self, file_path="", fmt="xlsx"):
        data = tablib.Dataset()
        cases = self.build()
        headers = []

        if len(cases) > 0:
            d = []
            for case in cases:
                values = []
                for cell in case:
                    for k, v in cell.items():
                        if k not in headers:
                            headers.append(k)
                        values.append(v)
                d.append(values)
                data.append(values)
            data.headers = headers
            export_data = data.export(fmt)

            with open(file_path, 'wb') as fw:
                fw.write(export_data)
        else:
            raise Exception(f"Case build failed, pls check fields in builder-schema.")



# -*- coding: utf-8 -*-
"""
# @Time    : 2022/3/4 15:32
# @Author  : bruce
# @desc    :
"""
import unittest
from unittest.case import _Outcome


class UnitHttpFrame(unittest.TestCase):

    __autoRequest__ = 'on'
    def __deepcopy__(self, memo):
        return self

    def initRequest(self, testmethod_name):
        pass

    def beforeRequest(self):
        pass

    def doRequest(self, request=None):
        pass

    def afterRequest(self):
        pass

    def run(self, result=None):
        orig_result = result
        if result is None:
            result = self.defaultTestResult()
            startTestRun = getattr(result, 'startTestRun', None)
            if startTestRun is not None:
                startTestRun()

        result.startTest(self)
        testMethod = getattr(self, self._testMethodName)
        if (getattr(self.__class__, "__unittest_skip__", False) or
            getattr(testMethod, "__unittest_skip__", False)):
            # If the class or method was skipped.
            try:
                skip_why = (getattr(self.__class__, '__unittest_skip_why__', '')
                            or getattr(testMethod, '__unittest_skip_why__', ''))
                self._addSkip(result, self, skip_why)
            finally:
                result.stopTest(self)
            return
        expecting_failure_method = getattr(testMethod,
                                           "__unittest_expecting_failure__", False)
        expecting_failure_class = getattr(self,
                                          "__unittest_expecting_failure__", False)
        expecting_failure = expecting_failure_class or expecting_failure_method
        outcome = _Outcome(result)
        try:
            self._outcome = outcome

            with outcome.testPartExecutor(self):
                self.setUp()
                self.beforeRequest()
            if outcome.success:
                outcome.expecting_failure = expecting_failure
                with outcome.testPartExecutor(self, isTest=True):
                    if self.__autoRequest__ == 'on':
                        self.doRequest()
                        self.afterRequest()
                    testMethod()
                outcome.expecting_failure = False
                with outcome.testPartExecutor(self):
                    self.tearDown()

            self.doCleanups()
            for test, reason in outcome.skipped:
                self._addSkip(result, test, reason)
            self._feedErrorsToResult(result, outcome.errors)
            if outcome.success:
                if expecting_failure:
                    if outcome.expectedFailure:
                        self._addExpectedFailure(result, outcome.expectedFailure)
                    else:
                        self._addUnexpectedSuccess(result)
                else:
                    result.addSuccess(self)
            return result
        finally:
            result.stopTest(self)
            if orig_result is None:
                stopTestRun = getattr(result, 'stopTestRun', None)
                if stopTestRun is not None:
                    stopTestRun()

            # explicitly break reference cycles:
            # outcome.errors -> frame -> outcome -> outcome.errors
            # outcome.expectedFailure -> frame -> outcome -> outcome.expectedFailure
            outcome.errors.clear()
            outcome.expectedFailure = None

            # clear the outcome, no more needed
            self._outcome = None



# -*- coding: utf-8 -*-
"""
# @Time    : 2022/6/9 19:30
# @Author  : bruce
# @desc    :
"""
import copy

from api_test_ez.ez.serialize.errors import ValidationError
from api_test_ez.ez.serialize.fields import BaseField, StringField, IntegerField


__all__ = ["ValidatorModel"]


class ModelMetaclass(type):

    def __new__(mcs, name, bases, attrs):
        super_new = super().__new__

        # If a base class just call super new
        if name == "ValidatorModel":
            return super_new(mcs, name, bases, attrs)

        # Discover any fields
        fields_mapping = {}

        for attr_name, attr_value in attrs.items():
            if issubclass(attr_value.__class__, BaseField) \
                    or attr_value.__class__.__base__.__name__ in __all__:   # is subclass of `ValidatorModel`
                fields_mapping.update({attr_name: attr_value})
                attrs.update({attr_name: attr_value})

        # Record all fields.
        attrs["__declared_fields__"] = fields_mapping

        return super_new(mcs, name, bases, attrs)


class ValidatorModel(metaclass=ModelMetaclass):

    def __init__(self, *args, **values):
        self.__fields_mapping__ = copy.deepcopy(getattr(self, "__declared_fields__"))
        if args:
            raise TypeError(
                "Instantiating a field with positional arguments is not "
                "supported. Please use `field_name=value` keyword arguments."
            )

        self.source_values = values

    def validate(self, data: dict, full_repr=True):
        """
        :param data:
        :param full_repr:   Determines the completeness of the returned result data.
                            Must be `True` when call internally.
        :return:
        """
        data.update(self.source_values)
        fields_mapping = copy.deepcopy(getattr(self, "__fields_mapping__"))

        for k, v in fields_mapping.items():
            try:
                # If `v` is `ValidateModel`, we validate it by itself.
                # We use `setattr` to set attributes, in order to trigger `__set__`.
                if isinstance(v, ValidatorModel):
                    setattr(self, k, v.validate(data[k]))
                else:
                    setattr(self, k, data[k])
                v = getattr(self, k)
                data.update({k: v})
            except KeyError:
                if getattr(v, 'required', None):
                    _error = ValidationError("Field is required but not provided.")
                    setattr(self, k, repr(_error))
                    data.update({k: repr(_error)})
        if full_repr:
            return data
        else:
            return getattr(self, "__fields_mapping__")

    def __str__(self):
        return f"<{self.__class__.__name__}> {getattr(self, '__fields_mapping__')}"

    __repr__ = __str__
